#include <Xm/DialogS.h>
#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Separator.h>
#include <Xm/Label.h>
#include <Xm/Text.h>

#include <X11/Xos.h>

#include <stdio.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <errno.h>

#include "askuser.h"


#define IS_DIRECTORY    1
#define IS_ERROR        2
#define IS_OKFILE       3
#define IS_NOFILE       4


static Widget gDialogParent;
static XtAppContext gDialogContext;
static Pixmap gDialogPixmap;

typedef struct
{
    int     answer;
    int     mustExist;
    char    *fileName;
} tFileReply;


void SetDialogParent(Widget parent)
{
    gDialogParent = parent;
}

void SetDialogAppContext(XtAppContext app)
{
    gDialogContext = app;
}

void SetDialogPixmap(Pixmap pixmap)
{
    gDialogPixmap = pixmap;
}

void AskUserCallback(Widget widget,
                     XtPointer clientData,
                     XtPointer callData)
{
    int     *answer;
    XmAnyCallbackStruct *cbs;

    answer = (int *)clientData;
    cbs = (XmAnyCallbackStruct *)callData;

    if (cbs->reason == XmCR_OK)
        *answer = YES;
    else if (cbs->reason == XmCR_CANCEL)
        *answer = NO;
}

char *AskUserForPrinter(void)
{
    static Widget   dialog;
    Arg             args[5];
    XmString        tmp;
    int             n;
    int             answer;
    char            *text;

    if (!dialog)
    {
        tmp = XmStringCreateLocalized("Print command: ");
        n = 0;
        XtSetArg(args[n], XmNselectionLabelString, tmp); n++;
        XtSetArg(args[n], XmNautoUnmanage, False); n++;
        XtSetArg(args[n], XmNuserData, 0); n++;

        dialog = XmCreatePromptDialog(gDialogParent,
                                      "printerQuery",
                                      args,
                                      n);
        XtVaSetValues(dialog,
                      XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
                      NULL);
        XmStringFree(tmp);
        XtVaSetValues(XmSelectionBoxGetChild(dialog,
                                             XmDIALOG_TEXT),
                      XmNvalue, "lp",
                      NULL);
                                             
        XtUnmanageChild(XmSelectionBoxGetChild(dialog,
                                             XmDIALOG_HELP_BUTTON));
        XtAddCallback(dialog, XmNokCallback, AskUserCallback, &answer);
        XtAddCallback(dialog, XmNcancelCallback, AskUserCallback, &answer);
    }
    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);

    answer = -1;

    while (answer == -1)
    {
        XtAppProcessEvent(gDialogContext, XtIMAll);
    }

    if (answer)
        XtVaGetValues(XmSelectionBoxGetChild(dialog, XmDIALOG_TEXT),
                      XmNvalue, &text,
                      NULL);
    else
        text = NULL;

    XtPopdown(XtParent(dialog));
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(gDialogParent);

    return text;
}


int AskUser(char *question,
            char *yesLabel, 
            char *noLabel,
            int   defaultAnswer)
{
    XmString    text, yes, no;
    static int  answer;
    static Widget dialog;

    if (!dialog)
    {
        dialog = XmCreateQuestionDialog(gDialogParent, "question", NULL, 0);
        XtVaSetValues(dialog,
                      XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
                      NULL);
        XtUnmanageChild(XmMessageBoxGetChild(dialog,
                                             XmDIALOG_HELP_BUTTON));
        XtAddCallback(dialog, XmNokCallback, AskUserCallback, &answer);
        XtAddCallback(dialog, XmNcancelCallback, AskUserCallback, &answer);
    }

    answer = -1;

    text = XmStringCreateLtoR(question, XmFONTLIST_DEFAULT_TAG);
    yes = XmStringCreateSimple(yesLabel);
    no = XmStringCreateSimple(noLabel);

    XtVaSetValues(dialog,
                  XmNmessageString, text,
                  XmNokLabelString, yes,
                  XmNcancelLabelString, no,
                  XmNdefaultButtonType,
                  defaultAnswer = YES? XmDIALOG_OK_BUTTON :
                  XmDIALOG_CANCEL_BUTTON,
                  NULL);

    XmStringFree(text);
    XmStringFree(yes);
    XmStringFree(no);

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);

    while (answer == -1)
    {
        XtAppProcessEvent(gDialogContext, XtIMAll);
    }

    XtPopdown(XtParent(dialog));
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(gDialogParent);

    return answer;
}

void TellUser(char *message)
{
    XmString    text, yes, no;
    static int  answer;
    static Widget dialog;

    if (!dialog)
    {
        dialog = XmCreateErrorDialog(gDialogParent, "question", NULL, 0);
        yes = XmStringCreateSimple("OK");
        no = XmStringCreateSimple("Cancel");
        XtVaSetValues(dialog,
                      XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
                      XmNokLabelString, yes,
                      XmNcancelLabelString, no,
                      NULL);
        XmStringFree(yes);
        XmStringFree(no);
        XtUnmanageChild(XmMessageBoxGetChild(dialog,
                                             XmDIALOG_HELP_BUTTON));
        XtSetSensitive(XmMessageBoxGetChild(dialog,
                                            XmDIALOG_CANCEL_BUTTON),
                                            False);
        XtAddCallback(dialog, XmNokCallback, AskUserCallback, &answer);
        XtAddCallback(dialog, XmNcancelCallback, AskUserCallback, &answer);
    }

    text = XmStringCreateLtoR(message, XmFONTLIST_DEFAULT_TAG);

    answer = -1;

    XtVaSetValues(dialog,
                  XmNmessageString, text,
                  XmNdefaultButtonType, XmDIALOG_OK_BUTTON,
                  NULL);

    XmStringFree(text);

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);

    while (answer == -1)
    {
        XtAppProcessEvent(gDialogContext, XtIMAll);
    }

    XtPopdown(XtParent(dialog));
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(gDialogParent);
}


void Error(char *fmt, ...)
{
    va_list     args;
    char        message[1024];

    va_start(args, fmt);

    vsprintf(message, fmt, args);
    TellUser(message);
}


int CheckFile(char *file)
{
    struct stat statBuffer;

    if (stat(file, &statBuffer) == -1)
        return IS_ERROR;
    else if ((statBuffer.st_mode & S_IFMT) == S_IFDIR)
        return IS_DIRECTORY;
    else if (!(statBuffer.st_mode & S_IFREG))
        return IS_NOFILE;
    else
        return IS_OKFILE;
}


void DontSelectFileCallback(Widget widget,
                        XtPointer clientData,
                        XtPointer callData)
{
    tFileReply *reply;

    reply = (tFileReply *)clientData;

    reply->answer = 1;
    reply->fileName = NULL;
}

void SelectFileCallback(Widget widget,
                        XtPointer clientData,
                        XtPointer callData)
{
    XmFileSelectionBoxCallbackStruct *cbs;
    tFileReply *reply;
    char *file, *fileName;

    reply = (tFileReply *)clientData;
    cbs = (XmFileSelectionBoxCallbackStruct *)callData;

    if (!XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &file))
        return;

    if (*file != '/')
    {
        char *dir;

        if (XmStringGetLtoR(cbs->dir, XmSTRING_DEFAULT_CHARSET, &dir))
        {
           fileName = XtMalloc(strlen(dir) + strlen(file) + 2);
           sprintf(fileName, "%s/%s", dir, file);
           XtFree(dir);
           XtFree(file);
        }
        fileName = file;
    }
    else
        fileName = file;

    switch (CheckFile(fileName))
    {
    case IS_DIRECTORY:
    {
        XmString    str;
        
        str = XmStringCreateSimple(fileName);
        XmFileSelectionDoSearch(widget, str);
        XmStringFree(str);
        XtFree(fileName);
        break;
    }
    case IS_NOFILE:
        break;
    case IS_OKFILE:
        reply->answer = 1;
        reply->fileName = fileName;
        break;
    case IS_ERROR:
        if (errno == ENOENT && !reply->mustExist)
        {
            reply->answer = 1;
            reply->fileName = fileName;
        }
        else
        {
            Error("%s", strerror(errno));
        }
        break;
    default:
        break;
    }
}

char *SelectFile(char *okButtonLabel, char mustExist)
{
    static Widget   dialog;
    tFileReply      reply;
    XmString        ok, no;

    if (!dialog)
    {
        dialog = XmCreateFileSelectionDialog(gDialogParent,
                                             "Files",
                                             NULL,
                                             0);
        no = XmStringCreateSimple("Cancel");
        XtVaSetValues(dialog,
                      XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
                      XmNcancelLabelString, no,
                      NULL);
        XmStringFree(no);
        XtSetSensitive(XmFileSelectionBoxGetChild(dialog,
                                                  XmDIALOG_HELP_BUTTON),
                       False);
        XtAddCallback(dialog, XmNokCallback, SelectFileCallback, &reply);
        XtAddCallback(dialog, XmNcancelCallback, DontSelectFileCallback,
                      &reply);
    }

    ok = XmStringCreateSimple(okButtonLabel);
    XtVaSetValues(dialog,
                  XmNokLabelString, ok,
                  NULL);

    reply.answer = 0;
    reply.mustExist = mustExist;

    XtManageChild(dialog);
    XtPopup(XtParent(dialog), XtGrabNone);
    
    while (reply.answer == 0)
    {
        XtAppProcessEvent(gDialogContext, XtIMAll);
    }

    XtPopdown(XtParent(dialog));
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(gDialogParent);

    return reply.fileName;
}
